/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000, 2001  Pan Development Team <pan@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <pan/base/argset.h>
#include <pan/base/debug.h>
#include <pan/base/file-grouplist.h>
#include <pan/base/file-headers.h>
#include <pan/base/gnksa.h>
#include <pan/base/log.h>
#include <pan/base/pan-config.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/pan-i18n.h>
#include <pan/base/serverlist.h>
#include <pan/base/status-item.h>

#include <pan/rules/rule-ui.h>

#include <pan/identities/identity.h>
#include <pan/identities/identity-manager.h>
#include <pan/identities/identity-ui.h>

#include <pan/article-actions.h>
#include <pan/article-find.h>
#include <pan/articlelist.h>
#include <pan/article-toolbar.h>
#include <pan/filter-ui.h>
#include <pan/filter-current-ui.h>
#include <pan/globals.h>
#include <pan/group-action.h>
#include <pan/group-ui.h>
#include <pan/gui.h>
#include <pan/gui-notebook.h>
#include <pan/gui-paned.h>
#include <pan/grouplist.h>
#include <pan/message-send.h>
#include <pan/message-window.h>
#include <pan/newsrc-ui.h>
#include <pan/pan.h>
#include <pan/prefs.h>
#include <pan/print.h>
#include <pan/server-ui.h>
#include <pan/status-item-view.h>
#include <pan/text.h>
#include <pan/task-bodies.h>
#include <pan/task-manager.h>
#include <pan/util.h>
#include <pan/queue.h>
#include <pan/dialogs/dialogs.h>

#include <pan/xpm/pan-pixbufs.h>

GtkTooltips * ttips = NULL;
GdkColormap * cmap = NULL;

static unsigned int gui_refresh_timer_id = 0;
static unsigned int gui_queue_timer_id = 0;
static double KBps = 0.0;
static GtkWidget * taskbar = NULL;

/**
***  Status Button
**/
typedef enum
{
	STATUS_OK,
	STATUS_ERROR,
	STATUS_QTY
}
StatusType;
typedef struct
{
	const guint8 * inline_txt;
	GdkPixbuf * pixbuf;
	GdkPixmap * pixmap;
	GdkBitmap * bitmap;
}
StatusIcon;
static StatusType _current_status;
static GtkWidget * _status_image = NULL;
static GtkWidget * _status_button;
static StatusIcon _status_icons[STATUS_QTY] = {
	{ icon_inform, NULL, NULL },
	{ icon_caution, NULL, NULL }
};

GtkAccelGroup * _main_accel_group = NULL;

#define VIEW_QTY 3
typedef struct {
	StatusItem * status_item;
	GtkWidget * view;
} ViewStruct;
static ViewStruct views[VIEW_QTY];

/**
*** private function prototypes
**/

static void exit_cb (void);
static int gui_page_get_nolock (void);
static GtkWidget* gui_page_get_pane_nolock (int page);
static gboolean gui_key_press_cb (GtkWidget*, GdkEventKey*, gpointer);
static int window_delete_event_cb (GtkWidget*, GdkEvent*, gpointer);
static int gui_refresh_timer_cb (gpointer);
static int gui_queue_timer_cb (gpointer);
static void update_menus_cb                      (gpointer, gpointer, gpointer);
static void log_entry_added_cb                   (gpointer, gpointer, gpointer);
static void status_item_active_changed_cb        (gpointer, gpointer, gpointer);
static void log_list_appended_cb                 (gpointer, gpointer, gpointer); 
static void text_fill_body_changed_cb            (gpointer, gpointer, gpointer); 
static void articles_changed_cb                  (gpointer, gpointer, gpointer);
static void turn_off_status_item_views_nolock    (void);
static void gui_set_connection_size (int size);

static void
server_import_newsrc_dialog_menu (void)
{
	Server * server = serverlist_get_active_server ();
	if (server != NULL)
		newsrc_import_dialog (server);
}
static void
server_export_newsrc_dialog_menu (void)
{
	Server * server = serverlist_get_active_server ();
	if (server != NULL)
		newsrc_export_dialog (server);
}

#define MAIN_MENU_ENTRIES_QTY 142
static GtkItemFactoryEntry main_menu_entries[MAIN_MENU_ENTRIES_QTY];
GtkItemFactory * _main_menu_factory;

/**
***  DIALOGS
**/

static void
widget_destroyed_cb (GtkObject * o, gpointer data)
{
	*((GtkWidget**)data) = NULL;
}

static GtkWidget * identity_dialog = NULL;

static void
identity_cb (void)
{
	if (identity_dialog == NULL)
	{
		identity_dialog = identity_dialog_new (GTK_WINDOW(Pan.window));
		g_signal_connect (GTK_OBJECT(identity_dialog), "destroy",
		                  G_CALLBACK(widget_destroyed_cb),
		                  &identity_dialog);
	}

	gtk_widget_show_all (identity_dialog);
}

static GtkWidget * filter_dialog = NULL;

static void
filter_custom_cb (void)
{
	if (filter_dialog == NULL)
	{
		filter_dialog = filter_dialog_new (GTK_WINDOW(Pan.window));
		g_signal_connect (GTK_OBJECT(filter_dialog), "destroy",
		                  G_CALLBACK(widget_destroyed_cb),
		                  &filter_dialog);
	}

	gtk_widget_show_all (filter_dialog);
}


static GtkWidget * rule_dialog = NULL;

static void
rules_cb (void)
{
	if (rule_dialog == NULL)
	{
		rule_dialog = rule_dialog_new (GTK_WINDOW(Pan.window));
		g_signal_connect (GTK_OBJECT(rule_dialog), "destroy",
		                  G_CALLBACK(widget_destroyed_cb),
		                  &rule_dialog);
	}

	gtk_widget_show_all (rule_dialog);
}

GtkWidget* create_rules_dialog (void);


static int current_layout = -1;

static void
navigate_read_on_select_cb (gpointer user_data, int action, GtkWidget * w)
{
	navigate_read_on_select = GTK_CHECK_MENU_ITEM(w)->active;
}

static void
gui_read_next (void)
{
	articlelist_read_next (navigate_read_on_select);
}


/* aein */

static void
pan_homepage_url (void)
{
	pan_url_show ("http://pan.rebelbase.com/");
}

static void
pan_bug_url (void)
{
	pan_url_show ("http://bugzilla.gnome.org/enter_bug.cgi?product=Pan");
}

static void
pan_manual_url (void)
{
	pan_url_show ("http://pan.rebelbase.com/manual/");
}

static void
send_feedback_cb (GtkWidget * w)
{
	char * from;
	char * msg_id;
	Group * sendlater;
	Article * a;
	Identity * id;
	debug_enter ("send_feedback_cb");

	/* who is sending this? */
	from = NULL;
       	id = identity_manager_get_default (ID_MAIL_DEFAULT);
	if (id != NULL)
		from = g_strdup_printf ("\"%s\" <%s>", id->author_real, id->author_addr);

       	msg_id = gnksa_generate_message_id ("rebelbase.com");
       	sendlater = serverlist_get_named_folder (PAN_SENDLATER);
       	a = article_new (sendlater);
	a->number = 1;

	article_init_header (a, HEADER_SUBJECT, "Pan " VERSION " Feedback", DO_CHUNK_SHARE);
	article_init_header (a, PAN_MAIL_TO, "pan@rebelbase.com", DO_CHUNK_SHARE);
	article_init_header (a, HEADER_MESSAGE_ID, msg_id, DO_CHUNK);
	if (is_nonempty_string(from))
		article_init_author_from_header (a, from);
	message_edit_window (a);

	g_free (msg_id);
	g_free (from);
	debug_exit ("send_feedback_cb");
}

static void
dialog_about_cb (void)
{
	dialog_about (Pan.window);
}



static void
thread_articles_cb (gpointer user_data, int action, GtkWidget * w)
{
	GtkCheckMenuItem * item = GTK_CHECK_MENU_ITEM(w);
	g_return_if_fail (item!=NULL);
	articlelist_set_threaded (item->active);
}

static void
rot13_cb (gpointer user_data, int action, GtkWidget * w)
{
	text_rot13_selected_text_nolock ();
}

static void
fill_body_cb (gpointer user_data, int action, GtkWidget * w)
{
	GtkCheckMenuItem * item = GTK_CHECK_MENU_ITEM(w);
	g_return_if_fail (item!=NULL);
	text_set_wrap (item->active);
}

static void
show_all_headers_cb (gpointer user_data, int action, GtkWidget * w)
{
	GtkCheckMenuItem * item = GTK_CHECK_MENU_ITEM(w);
	g_return_if_fail (item!=NULL);
	text_set_show_all_headers (item->active);
}

static void
mute_quoted_text_cb (gpointer user_data, int action, GtkWidget * w)
{
	GtkCheckMenuItem * item = GTK_CHECK_MENU_ITEM(w);
	g_return_if_fail (item!=NULL);
	text_set_mute_quoted (item->active);
}

static void
use_fixed_font_cb (gpointer user_data, int action, GtkWidget * w)
{
	GtkCheckMenuItem * item = GTK_CHECK_MENU_ITEM(w);
	g_return_if_fail (item!=NULL);
	text_use_fixed_font = item->active;
	text_set_font ();
}

static void
collapse_group_names_cb (gpointer user_data, int action, GtkWidget * w)
{
	GtkCheckMenuItem * item = GTK_CHECK_MENU_ITEM(w);
	if (collapse_group_names != item->active) {
		collapse_group_names = item->active;
		grouplist_refresh_nolock ();
	}
}

static void
set_current_status_nolock (StatusType status)
{
	g_return_if_fail (status==STATUS_OK || status==STATUS_ERROR);

	_current_status = status;
	gtk_image_set_from_pixmap (GTK_IMAGE(_status_image),
	                           _status_icons[_current_status].pixmap,
	                           _status_icons[_current_status].bitmap);
}

static int
log_button_clicked_idle (gpointer data)
{
	pan_lock ();
	set_current_status_nolock (STATUS_OK);
	pan_unlock ();

	return 0;
}

static void
log_button_clicked (void)
{
	gui_queue_add (log_button_clicked_idle, NULL);

	dialog_log_viewer ();
}

static void
toggle_pane_visibility (gpointer user_data, int action, GtkWidget * w)
{
	GtkCheckMenuItem * item = GTK_CHECK_MENU_ITEM (w);

	if (current_layout == GUI_PANED)
	{
		gboolean * show_pane = NULL;
		const guint page = GPOINTER_TO_UINT(user_data);

		switch (page) {
			case GROUPS_PAGE: show_pane = &show_group_pane; break;
			case ARTICLELIST_PAGE: show_pane = &show_header_pane; break;
			case MESSAGE_PAGE: show_pane = &show_article_pane; break;
			default: pan_warn_if_reached ();
		}

		if (show_pane!=NULL && *show_pane!=item->active) {
			*show_pane = item->active;
			gui_layout_set_mode (current_layout);
		}
	}
}

static void
zoom_page_cb (gpointer menu, gpointer data)
{
	const guint page = GPOINTER_TO_UINT(data);
       	gui_layout_set_mode (GUI_NOTEBOOK);
	gui_page_set (page);
}

static void
zoom_cb (void)
{
	const int page = gui_page_get_nolock ();
       	gui_layout_set_mode (current_layout==GUI_PANED ? GUI_NOTEBOOK : GUI_PANED);
	gui_page_set (page);
}


static void
mute_quoted_text_changed_cb (gpointer data, gpointer call, gpointer client)
{
	menu_set_checked (_main_menu_factory, "/View/Mute Quoted Text", call!=NULL);
}

static void
show_all_headers_changed_cb (gpointer data, gpointer call, gpointer client)
{
	menu_set_checked (_main_menu_factory, "/View/Show All Headers in Article Pane", call!=NULL);
}



/** 
 ** global variables
 **/

static GtkWidget * appbar = NULL;
static GtkWidget * queue_qty_label = NULL;
static GtkWidget * queue_qty_button = NULL;
static GtkWidget * connection_qty_label = NULL;
static GtkWidget * connection_qty_button = NULL;

guint context_id;

GtkWidget * groups_vbox;
GtkWidget * articlelist_ctree;
GtkWidget * text_box;

static void
connection_button_clicked (GtkButton *button,
                           gpointer user_data)
{
	prefs_spawn_to_news_connections ();
}

static void
queue_button_clicked (GtkButton *button,
                      gpointer user_data)
{
	task_manager_spawn ();
}


static void
gui_shutdown (void)
{
	int i;
	char * pch;
	GtkWidget * w[3];
	debug_enter ("gui_shutdown");

	/**
	***  Stop updating the UI
	**/

	pan_callback_remove (status_item_get_active_callback(),
	                     status_item_active_changed_cb, NULL);
	turn_off_status_item_views_nolock ();

	/**
	***  Stop updating the rest of the UI
	**/

	pan_callback_remove (article_get_articles_changed_callback(),
	                     articles_changed_cb, NULL);
	pan_callback_remove (log_get_entry_added_callback(),
	                     log_entry_added_cb, NULL);
	pan_callback_remove (log_get_entry_added_callback(),
	                     log_list_appended_cb, NULL);
	pan_callback_remove (text_get_fill_body_changed_callback(),
	                     text_fill_body_changed_cb, NULL);
	pan_callback_remove (text_get_show_all_headers_changed_callback(),
	                     show_all_headers_changed_cb, NULL);
	pan_callback_remove (text_get_mute_quoted_changed_callback(),
	                     mute_quoted_text_changed_cb, NULL);
	pan_callback_remove (current_article_changed,
	                     update_menus_cb, NULL);
	pan_callback_remove (grouplist_group_selection_changed,
	                     update_menus_cb, NULL);
	pan_callback_remove (grouplist_server_set_callback,
	                     update_menus_cb, NULL);
	pan_callback_remove (articlelist_get_group_changed_callback(),
	                     update_menus_cb, NULL);
	pan_callback_remove (articlelist_selection_changed,
	                     update_menus_cb, NULL);

	g_source_remove (gui_refresh_timer_id);
	gui_refresh_timer_id = 0;
	g_source_remove (gui_queue_timer_id);
	gui_queue_timer_id = 0;

	/**
	***  Save settings
	**/

	pan_config_push_prefix ("/Pan/Display/");
	pan_config_set_bool ("show_group_pane", show_group_pane);
	pan_config_set_bool ("header_pane_is_threaded", header_pane_is_threaded);
	pan_config_set_bool ("show_header_pane", show_header_pane);
	pan_config_set_bool ("show_article_pane", show_article_pane);
	pan_config_set_bool ("navigate_read_on_select", navigate_read_on_select);
	pan_config_set_bool ("collapse_group_names", collapse_group_names);
	pan_config_set_bool ("do_wrap", text_get_wrap());
	pan_config_set_bool ("use_fixed_font", text_use_fixed_font);
	pan_config_set_bool ("Show_All_Headers_In_Body", text_get_show_all_headers());
	pan_config_set_bool ("Mute_Quoted_text", text_get_mute_quoted());

	pan_config_pop_prefix ();

	pan_config_set_int ("/Pan/State/viewmode", Pan.viewmode);

	grouplist_shutdown_module ();

	switch (Pan.viewmode)
	{
		case GUI_NOTEBOOK:
			pan_config_set_int ("/Pan/State/page", gui_page_get_nolock ());
			break;

		case GUI_PANED:
			pch = layout_str + 1;
			for (i=0; i<=2; ++i, ++pch)
			{
				if (tolower(*pch)=='g')
					w[i] = groups_vbox;
				else if (tolower(*pch)=='t')
					w[i] = articlelist_ctree;
				else
					w[i] = text_box;
			}
			switch(*layout_str)
			{
				case '2':
					pan_config_set_int ("/Pan/Geometry/vpaned", w[0]->allocation.height);
					pan_config_set_int ("/Pan/Geometry/hpaned", w[1]->allocation.width);
					break;
				case '4':
					pan_config_set_int ("/Pan/Geometry/vpaned", w[1]->allocation.height);
					pan_config_set_int ("/Pan/Geometry/hpaned", w[0]->allocation.width);
					break;
				case '5':
					pan_config_set_int ("/Pan/Geometry/vpaned", w[0]->allocation.height);
					pan_config_set_int ("/Pan/Geometry/hpaned", w[0]->allocation.height + w[1]->allocation.height);
					break;
				case '1':
				case '3':
				default:
					pan_config_set_int ("/Pan/Geometry/vpaned", w[0]->allocation.height);
					pan_config_set_int ("/Pan/Geometry/hpaned", w[0]->allocation.width);
					break;
			}
	}

	gui_save_column_widths (Pan.group_tree, "group");
	gui_save_column_widths (Pan.article_ctree, "thread_pane");
	gui_save_window_size (Pan.window, "main_window");
	pan_config_sync();

	/**
	***  Clear out the UI pieces
	**/

	gtk_widget_unref (groups_vbox);
	gtk_widget_unref (articlelist_ctree);
	gtk_widget_unref (text_box);

	articlelist_set_group_nolock (NULL);
	text_clear_nolock ();
	debug_exit ("gui_shutdown");
}

static GtkWidget*
gui_create_appbar (GtkWidget* window)
{
	int i;
	GtkWidget * w;
	GtkWidget * frame;

	appbar = gtk_hbox_new (FALSE, GUI_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER(appbar), GUI_PAD_SMALL);

	/* connection status */
	connection_qty_label = gtk_label_new ("");
	w = connection_qty_button = gtk_button_new();
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w, _("Open the Connection Manager"), "");
	gtk_button_set_relief (GTK_BUTTON(w), GTK_RELIEF_NONE);
	g_signal_connect (GTK_OBJECT(w), "clicked", G_CALLBACK(connection_button_clicked), NULL);
	gtk_container_add (GTK_CONTAINER(w), connection_qty_label);
	frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_IN);
	gtk_container_add (GTK_CONTAINER(frame), w);
	gtk_box_pack_start (GTK_BOX(appbar), frame, FALSE, FALSE, 0);

	/* task status */
	queue_qty_label = gtk_label_new ("");
	w = queue_qty_button = gtk_button_new();
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w, _("Open the Task Manager"), "");
	gtk_button_set_relief (GTK_BUTTON(w), GTK_RELIEF_NONE);
	g_signal_connect (GTK_OBJECT(w), "clicked", G_CALLBACK(queue_button_clicked), NULL);
	gtk_container_add (GTK_CONTAINER(w), queue_qty_label);
	frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_IN);
	gtk_container_add (GTK_CONTAINER(frame), w);
	gtk_box_pack_start (GTK_BOX(appbar), frame, FALSE, FALSE, 0);

	/* status item views */
	taskbar = gtk_table_new (1, VIEW_QTY, TRUE);
	for (i=0; i<VIEW_QTY; ++i) {
		GtkWidget * w = status_item_view_new ();
                views[i].view = w;
		gtk_table_attach (GTK_TABLE(taskbar), w, i, i+1, 0, 1, ~0, ~0, 0, 0);
	}
	gtk_box_pack_start (GTK_BOX(appbar), taskbar, TRUE, TRUE, 0);

	/* status icons */
	for (i=0; i<STATUS_QTY; ++i) {
		_status_icons[i].pixbuf = gdk_pixbuf_new_from_inline (-1, _status_icons[i].inline_txt,
		                                                      FALSE, NULL);
		gdk_pixbuf_render_pixmap_and_mask_for_colormap (_status_icons[i].pixbuf, cmap,
		                                                &_status_icons[i].pixmap,
		                                                &_status_icons[i].bitmap, 128);
	}

	/* status button */
	_status_button = w = gtk_button_new ();
	_status_image = gtk_image_new ();
	gtk_container_add (GTK_CONTAINER(w), _status_image);
	set_current_status_nolock (STATUS_OK);
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w, _("Open the Status Log"), "");
	gtk_button_set_relief (GTK_BUTTON(w), GTK_RELIEF_NONE);
	g_signal_connect (GTK_OBJECT(w), "clicked", G_CALLBACK(log_button_clicked), NULL);
	gtk_box_pack_start (GTK_BOX(appbar), w, FALSE, FALSE, 0);

	gui_set_queue_size (0u, 0u);
	gui_set_connection_size (0);

	return appbar;
}

/*---[ gui_shutdown ]-------------------------------------------------
 * save geometry states into config files
 *--------------------------------------------------------------------*/
static guint
warn_about_tasks_running (void)
{
	guint num_tasks = queue_get_running_task_count ();
	int retval = TRUE;

	if (num_tasks != 0)
	{

		char * msg;
		int response = 0;
		GtkWidget * dialog;

		if (num_tasks == 1)
			msg = g_strdup( _("1 task still active or queued.\nAre you sure you want to exit Pan?"));
		else
			msg = g_strdup_printf (_("%d tasks still active or queued.\nAre you sure you want to exit Pan?"), num_tasks);

		dialog = gtk_message_dialog_new (GTK_WINDOW(Pan.window),
		                                 GTK_DIALOG_DESTROY_WITH_PARENT,
		                                 GTK_MESSAGE_QUESTION,
		                                 GTK_BUTTONS_YES_NO,
						 msg);

		g_free (msg);

		response = gtk_dialog_run (GTK_DIALOG(dialog));

		retval = response == GTK_RESPONSE_YES;
	}

	return retval;
}

static void
exit_cb (void)
{
	if (warn_about_tasks_running())
	{
		gui_shutdown ();
		gtk_object_destroy (GTK_OBJECT(Pan.window));
	}
}

static int
window_delete_event_cb (GtkWidget *widget,
			GdkEvent *event,
			gpointer  data)
{
	if (warn_about_tasks_running())
	{
		gui_shutdown ();
		return FALSE; /* causes "destroy" signal to be fired */
	}

	return TRUE;
}

static void
text_fill_body_changed_cb (gpointer call_obj,
                           gpointer call_arg,
                           gpointer user_data)
{
	menu_set_checked (_main_menu_factory, "/View/Fill & Wrap Article Body", call_arg!=NULL);
}

static int
log_list_appended_idle (gpointer data)
{
	pan_lock ();
	set_current_status_nolock (STATUS_ERROR);
	pan_unlock ();

	return 0;
}

static void
log_list_appended_cb (gpointer call_obj, gpointer call_arg, gpointer user_data)
{
	if (((LogEntry*)call_obj)->severity== LOG_ERROR)
		gui_queue_add (log_list_appended_idle, NULL);
}


/***
****
***/

static int
gui_set_title_idle (gpointer data)
{
	Server * server;
	const Group * group = GROUP(data);
	GString * s = g_string_new (NULL);
	const char * servername = NULL;
	const char * groupname = NULL;
	debug_enter ("gui_set_title_idle");

	/* look for active server & group */
	server = serverlist_get_active_server ();
	if (server != NULL)
		servername = server_get_name (server);
	if (group != NULL)
		groupname = group_get_readable_name (group);

	/* build the title */
	g_string_sprintf (s, "Pan %s ", VERSION);
	if (servername != NULL)
		g_string_sprintfa (s, " - %s", servername);
	if (groupname != NULL)
		g_string_sprintfa (s, " - %s - %d/%d",
		                   groupname,
		                   group->article_qty-group->article_read_qty,
		                   group->article_qty);

	/* set the title in the UI */
	pan_lock ();
	gtk_window_set_title (GTK_WINDOW (Pan.window), s->str);
	pan_unlock ();

	/* cleanup */
	g_string_free (s, TRUE);

	debug_exit ("gui_set_title_idle");
	return 0;
}

static void
gui_set_title (const Group * group)
{
	gui_queue_add (gui_set_title_idle, (gpointer)group);
}

static void
articles_changed_cb (gpointer call_obj, gpointer call_arg, gpointer user_data)
{
	Group * agroup = articlelist_get_group ();
	ArticleChangeEvent * e = (ArticleChangeEvent*) call_obj;
	if (agroup == e->group)
		gui_set_title (agroup);
}

/***
****
***/

static int
update_menus_idle (gpointer data)
{
	//int i;
	const Article * article;
	gboolean have_article;
	gboolean have_current_article;
	const Group * alist_group;
	const Group * glist_group;
	gboolean have_group;
	gboolean have_alist_folder;
	gboolean have_glist_folder;
	const Group * sendlater;
	gboolean sendlater_empty;
	gboolean posted_here;
	gboolean is_paned = current_layout == GUI_PANED;
	GtkItemFactory * gif = _main_menu_factory;
	debug_enter ("update_menus_idle");

	article = articlelist_get_selected_article_nolock ();
	have_article = article!=NULL;
	have_current_article = get_current_article()!=NULL;
	posted_here = article!=NULL;

	alist_group = articlelist_get_group ();
	glist_group = grouplist_get_selected_group();
	have_group = alist_group!=NULL || glist_group!=NULL;
	have_alist_folder = alist_group!=NULL && group_is_folder(alist_group);
	have_glist_folder = glist_group!=NULL && group_is_folder(glist_group);

	sendlater = serverlist_get_named_folder (PAN_SENDLATER);
	sendlater_empty = sendlater!=NULL && !sendlater->article_qty;

	pan_lock ();

	/* update view menu */
	menu_set_sensitive (gif, "/View/Show Group Pane", is_paned);
	menu_set_sensitive (gif, "/View/Show Header Pane", is_paned);
	menu_set_sensitive (gif, "/View/Show Article Pane", is_paned);

	/* post menu */
	menu_set_sensitive (gif, "/Post/Post to Newsgroup", TRUE);
	menu_set_sensitive (gif, "/Post/Follow-Up to Newsgroup", have_current_article);
	menu_set_sensitive (gif, "/Post/Reply by E-Mail", have_current_article);
	menu_set_sensitive (gif, "/Post/Follow-Up and Reply", have_current_article);
	menu_set_sensitive (gif, "/Post/Forward Article by E-Mail", have_current_article);
	menu_set_sensitive (gif, "/Post/Send Pending Messages", !sendlater_empty);

	/* navigate menu */
	menu_set_sensitive (gif, "/Go/Load Articles When Selected:", TRUE);
	menu_set_sensitive (gif, "/Go/Next Unread Article", have_group);
	menu_set_sensitive (gif, "/Go/Next Article", have_group);
	menu_set_sensitive (gif, "/Go/Next Unread Thread", have_group);
	menu_set_sensitive (gif, "/Go/Next Thread", have_group);
	menu_set_sensitive (gif, "/Go/Previous Article", have_group);
	menu_set_sensitive (gif, "/Go/Previous Thread", have_group);
	menu_set_sensitive (gif, "/Go/Previous Article Read", have_group && articlelist_has_prev_read());
	menu_set_sensitive (gif, "/Go/Parent Article", have_group);
	menu_set_sensitive (gif, "/Go/Top of Thread", have_group);
	menu_set_sensitive (gif, "/Go/Next Unread Group", TRUE);
	menu_set_sensitive (gif, "/Go/Next Group", TRUE);

	/* update group menu */
	menu_set_sensitive (gif, "/Newsgroups/Mark Group Read", have_group);
	menu_set_sensitive (gif, "/Newsgroups/Delete Group's Articles", have_group);
	menu_set_sensitive (gif, "/Newsgroups/Get New Headers", have_group);
	menu_set_sensitive (gif, "/Newsgroups/Get New Headers and Bodies", have_group);
	menu_set_sensitive (gif, "/Newsgroups/More Download Options...", have_group);
	menu_set_sensitive (gif, "/Newsgroups/Refresh Article Counts", have_group);
	menu_set_sensitive (gif, "/Newsgroups/Subscribe", have_group);
	menu_set_sensitive (gif, "/Newsgroups/Unsubscribe", have_group);
	menu_set_sensitive (gif, "/Newsgroups/Group Properties...", have_group);
	menu_set_sensitive (gif, "/Newsgroups/Delete Group", have_group);

	/* update article_menus */
	menu_set_sensitive (gif, "/Articles/Mark Read", have_article);
	menu_set_sensitive (gif, "/Articles/Mark Unread", have_article);
	menu_set_sensitive (gif, "/Articles/Download Flagged", TRUE);
	menu_set_sensitive (gif, "/Articles/Flag", have_article);
	menu_set_sensitive (gif, "/Articles/Unflag", have_article);
	menu_set_sensitive (gif, "/Articles/Download", have_article);
	menu_set_sensitive (gif, "/Articles/Watch Thread", have_article);
	menu_set_sensitive (gif, "/Articles/Ignore Thread", have_article);
	menu_set_sensitive (gif, "/Articles/Add to Bozo or Spam Filter...", have_article);
	menu_set_sensitive (gif, "/Articles/Copy to Folder", have_article);
	menu_set_sensitive (gif, "/Articles/Cancel...", have_article);
	menu_set_sensitive (gif, "/Articles/Supersede...", have_article);
	menu_set_sensitive (gif, "/Articles/Delete",  have_article);

	/* update file menu */
	menu_set_sensitive (gif, "/File/Save Article As...",  have_article);
	menu_set_sensitive (gif, "/File/Save Article Attachments",  have_article);
	menu_set_sensitive (gif, "/File/Print Article...",  have_article);
	menu_set_sensitive (gif, "/File/New Folder...",  TRUE);
	menu_set_sensitive (gif, "/File/Delete Folder...",  have_glist_folder && strncmp(glist_group->name, "pan.", 4));
	menu_set_sensitive (gif, "/File/Edit Article in Folder",  have_alist_folder && have_article);
	menu_set_sensitive (gif, "/File/Exit", TRUE);

	pan_unlock ();

	/* update the title */
	gui_set_title_idle ((gpointer)alist_group);

	debug_exit ("update_menus_idle");
	return 0;
}

static void
update_menus_cb (gpointer call_object, gpointer call_arg, gpointer user_data)
{
	gui_queue_add (update_menus_idle, NULL);
}

static void
log_entry_added_cb (gpointer entry_data, gpointer unused1, gpointer unused2)
{
	LogEntry* e = (LogEntry*) entry_data;
	if (e->severity & LOG_URGENT)
	{
		if (e->severity & LOG_ERROR)
			pan_error_dialog (e->message);
		else /* LOG_INFO */
			pan_info_dialog (e->message);
	}
}

static char*
menu_translate(const char* path, gpointer data)
{
	return gettext (path);
}

/****
*****
*****   GUI CONSTRUCT
*****
****/

void
gui_construct (const char * geometry)
{
	GtkWidget * w;
	GtkWidget * vbox;
	GtkWidget * menubar;
	debug_enter ("gui_construct");

	ttips = gtk_tooltips_new();
        cmap = gdk_colormap_get_system ();

	/**
	***  Make the Main Window
	**/

	/* create the main window */
	Pan.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW(Pan.window), "Pan");
	if (!gui_restore_window_size (Pan.window, "main_window"))
		gtk_window_set_default_size (GTK_WINDOW(Pan.window), 820, 700);
	g_signal_connect (GTK_OBJECT(Pan.window), "delete_event", 
	                  G_CALLBACK (window_delete_event_cb), NULL);
	g_signal_connect (GTK_OBJECT(Pan.window), "destroy",
	                  G_CALLBACK (pan_shutdown), NULL);

	/* set the window icon */
	if (1) {
		GdkPixbuf * icon = gdk_pixbuf_new_from_inline (-1, icon_pan, FALSE, NULL);
		GList * l = g_list_append (NULL, icon);
		gtk_window_set_default_icon_list (l);
		g_list_free (l);
	}

	/* create the table */
	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER(Pan.window), vbox);

	/* create the menubar */
	_main_accel_group = gtk_accel_group_new ();
	gtk_window_add_accel_group (GTK_WINDOW(Pan.window), _main_accel_group);
	g_object_unref (_main_accel_group);
	_main_menu_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", _main_accel_group);
	gtk_item_factory_set_translate_func (_main_menu_factory, menu_translate, NULL, NULL);
	gtk_item_factory_create_items (_main_menu_factory, MAIN_MENU_ENTRIES_QTY, main_menu_entries, NULL);
	menubar = gtk_item_factory_get_widget (_main_menu_factory, "<main>");
	gtk_box_pack_start (GTK_BOX(vbox), menubar, FALSE, FALSE, 0);

	/* create the workarea */
	Pan.workarea = gtk_vbox_new (TRUE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), Pan.workarea, TRUE, TRUE, 0);

	/* create the statusbar */
	memset (views, '\0', sizeof(ViewStruct)*VIEW_QTY);
	w = gui_create_appbar (Pan.window);
	gtk_box_pack_start (GTK_BOX(vbox), w, FALSE, FALSE, 0);

	/**
	***  Make the Workarea Widgetry
	**/	

	/* group pane */
	groups_vbox = gtk_vbox_new (FALSE, 0);
	gtk_widget_set_usize (GTK_WIDGET (groups_vbox), 320, -1);
	gtk_box_pack_start (GTK_BOX (groups_vbox), GTK_WIDGET(grouplist_create()), TRUE, TRUE, 1);
	g_signal_connect (GTK_OBJECT (Pan.group_tree), "key_press_event",
			  G_CALLBACK (gui_key_press_cb),  NULL);

	/* article pane */
	articlelist_ctree = create_articlelist_ctree ();
        g_signal_connect (GTK_OBJECT (Pan.article_ctree), "key_press_event",
                          G_CALLBACK (gui_key_press_cb), NULL);

	/* text pane */
	text_box = text_create ();
        g_signal_connect (GTK_OBJECT (Pan.text), "key_press_event",
                          G_CALLBACK (gui_key_press_cb), NULL);

	/* set the workarea layout */
	gtk_widget_ref (groups_vbox);
	gtk_widget_ref (articlelist_ctree);
	gtk_widget_ref (text_box);
        gui_layout_set_mode (pan_config_get_int ( "/Pan/State/viewmode=1" ));

	/* listen for changes to the articlelist active group
	   so that we can sensitize/desensitize menus accordingly */
	update_menus_cb (NULL, NULL, NULL);

	pan_callback_add (article_get_articles_changed_callback(), articles_changed_cb, NULL);
	pan_callback_add (log_get_entry_added_callback(), log_list_appended_cb, NULL);
	pan_callback_add (status_item_get_active_callback(), status_item_active_changed_cb, NULL);
	pan_callback_add (text_get_fill_body_changed_callback(), text_fill_body_changed_cb, NULL);
	pan_callback_add (text_get_show_all_headers_changed_callback(), show_all_headers_changed_cb, NULL);
	pan_callback_add (text_get_mute_quoted_changed_callback(), mute_quoted_text_changed_cb, NULL);
	pan_callback_add (current_article_changed, update_menus_cb, NULL);
	pan_callback_add (grouplist_group_selection_changed, update_menus_cb, NULL);
	pan_callback_add (grouplist_server_set_callback, update_menus_cb, NULL);
	pan_callback_add (articlelist_get_group_changed_callback(), update_menus_cb, NULL);
	pan_callback_add (articlelist_selection_changed, update_menus_cb, NULL );
	pan_callback_add (log_get_entry_added_callback(), log_entry_added_cb, NULL);

	menu_set_checked (_main_menu_factory, "/View/Collapse Names in Group Pane", collapse_group_names);
	menu_set_checked (_main_menu_factory, "/View/Fill & Wrap Article Body", text_get_wrap());
	menu_set_checked (_main_menu_factory, "/View/Show All Headers in Article Pane", text_get_show_all_headers());
	menu_set_checked (_main_menu_factory, "/View/Use Monospace Font in Article Pane", text_use_fixed_font);
	menu_set_checked (_main_menu_factory, "/View/Thread Header Pane", header_pane_is_threaded);
	menu_set_checked (_main_menu_factory, "/Go/Load Articles When Selected:", navigate_read_on_select);
	menu_set_checked (_main_menu_factory, "/View/Show Group Pane", show_group_pane);
	menu_set_checked (_main_menu_factory, "/View/Show Header Pane", show_header_pane);
	menu_set_checked (_main_menu_factory, "/View/Show Article Pane", show_article_pane);
	menu_set_checked (_main_menu_factory, "/Go/Load Articles When Selected:", navigate_read_on_select);

	text_set_show_all_headers (pan_config_get_bool("/Pan/Display/Show_All_Headers_In_Body=false"));
	text_set_mute_quoted (pan_config_get_bool("/Pan/Display/Mute_Quoted_Text=false"));
	gui_set_title (NULL);
	gui_refresh_timer_id = pan_timeout_add (3000, gui_refresh_timer_cb, NULL); 
	gui_queue_timer_id = pan_timeout_add (160, gui_queue_timer_cb, NULL);
	gtk_widget_show_all (Pan.window);
	server_menu_update ();

	debug_exit ("gui_construct");
}

/****
*****
*****   GUI LAYOUT  (NOTEBOOK, PANED)
*****
****/

void
gui_layout_clear_workarea (void)
{
	/* clear out the workarea */
	gtk_widget_hide_all (GTK_WIDGET(Pan.workarea));
	if (groups_vbox->parent != NULL)
		gtk_container_remove (GTK_CONTAINER(groups_vbox->parent), groups_vbox);
	if (articlelist_ctree->parent != NULL)
		gtk_container_remove (GTK_CONTAINER(articlelist_ctree->parent), articlelist_ctree);
	if (text_box->parent != NULL)
		gtk_container_remove (GTK_CONTAINER(text_box->parent), text_box);

	/* make a new UI */
	if (1) {
		GList * children = gtk_container_children (GTK_CONTAINER(Pan.workarea));
		GList * l;
		for (l=children; l!=NULL; l=l->next)
			gtk_container_remove (GTK_CONTAINER(Pan.workarea), GTK_WIDGET(l->data));
		g_list_free (l);
	}
}

void
gui_layout_set_mode (int new_layout)
{
	switch (new_layout) {
		case GUI_NOTEBOOK:
			gui_notebook_construct ();
			break;
		case GUI_PANED:
			gui_paned_construct ();
			break;
		default:
			pan_warn_if_reached ();
			gui_paned_construct ();
			break;
	}

	current_layout = new_layout;
	update_menus_cb (NULL, NULL, NULL);
}

static int
gui_page_set_idle (gpointer page_gpointer)
{
	int page = GPOINTER_TO_INT (page_gpointer);
	GtkWidget * focus_pane;
	debug_enter ("gui_page_set_idle");

	pan_lock ();
	focus_pane = gui_page_get_pane_nolock (page);
	switch (Pan.viewmode)
	{
		case GUI_PANED:
			/* small tweak: it's not very useful to focus the message pane */
			if (page==ARTICLELIST_PAGE || page==MESSAGE_PAGE)
				focus_pane = gui_page_get_pane_nolock (ARTICLELIST_PAGE);
		    	gui_paned_page_set_nolock (page, focus_pane);
			break;

		case GUI_NOTEBOOK:
			gui_notebook_page_set_nolock (page, focus_pane);
			break;

		default:
			pan_warn_if_reached ();
			break;
	}
	pan_unlock ();

	debug_exit ("gui_page_set_idle");
	return 0;
}
void
gui_page_set (int page)
{
	gui_queue_add (gui_page_set_idle, GINT_TO_POINTER(page));
}

static int
gui_page_get_nolock (void)
{
	int retval = 0;

	switch (Pan.viewmode) {
		case GUI_PANED:    retval = gui_paned_get_current_pane_nolock();    break;
		case GUI_NOTEBOOK: retval = gui_notebook_get_current_pane_nolock(); break;
		default: pan_warn_if_reached(); break;
	}

	return retval;
}

static GtkWidget*
gui_page_get_pane_nolock (int page)
{
	GtkWidget * retval = NULL;

	switch (page) {
		case GROUPS_PAGE: retval = Pan.group_tree; break;
		case ARTICLELIST_PAGE: retval = Pan.article_ctree; break;
		case MESSAGE_PAGE: retval = Pan.text; break;
		default: pan_warn_if_reached ();
	}

	return retval;
}

static void
gui_page_activate_nolock (int page)
{
	switch (page)
	{
		case GROUPS_PAGE:
		{
			Group * g = grouplist_get_selected_group ();
			if (g != NULL)
				articlelist_set_group_nolock (g);
			break;
		}
		case ARTICLELIST_PAGE:
		{
			GPtrArray * articles;

			gui_queue_add ((GSourceFunc) articlelist_read_selected, NULL);

			/* also download any others that are selected */
			articles = articlelist_get_selected_articles_nolock ();
			if (articles->len > 1) {
				Group * g = ARTICLE(g_ptr_array_index(articles,0))->group;
				g_ptr_array_remove_index (articles, 0);
				queue_add (TASK(task_bodies_new (g, articles)));
			}
			g_ptr_array_free (articles, TRUE);

			break;
		}
		case MESSAGE_PAGE:
			break;
		default:
			pan_warn_if_reached ();
			break;
	}
}

static void
gui_page_change_nolock (int change)
{
	int page = gui_page_get_nolock () + change;

	if (0<=page && page<=2)
		gui_page_set (page);
}


/*---[ gui_key_press_cb ]---------------------------------------------
 * key press callback, when a key is pressed in the group list, article
 * list, or in the message view.  This allows better keyboard navigation
 * and the potential for new key bindings to imitate Ag*nt.
 *--------------------------------------------------------------------*/
static gboolean
gui_key_press_cb (GtkWidget      * widget,
                  GdkEventKey    * event,
                  gpointer         data)
{
	gboolean retval = FALSE;

	switch (event->keyval)
	{
		case GDK_Q: case GDK_q:
			gui_page_change_nolock (-1);
			retval = TRUE;
			break;

		case GDK_Delete:
			article_action_delete_selected_articles ();
			retval = TRUE;
			break;

		case GDK_Return:
			gui_page_activate_nolock (gui_page_get_nolock ());
			gui_page_change_nolock (1);
			retval = TRUE;
			break;

		case GDK_space: case GDK_KP_Space:
			text_read_more ();
			retval = TRUE;
			break;

		default:
			break;
	}

	return retval;
}

/**
***
**/

void
gui_restore_column_widths (GtkWidget     * clist,
                           const char   * type)
{
	if (GTK_IS_CLIST (clist))
	{
		int i;
		GtkCList * list = GTK_CLIST (clist);
		const int cols = list->columns;
		int * widths = g_new (int, cols);
		char buf[1024];
		debug_enter ("gui_restore_column_widths");

		/* get width from config... */
		for (i=0; i!=cols; ++i) {
			g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_column_%d", type, i);
			widths[i] = pan_config_get_int (buf);
		}

		/* set ui.. */
		pan_lock();
		for (i=0; i!=cols; ++i)
			if (widths[i] != 0)
				gtk_clist_set_column_width (list, i, widths[i]);
		pan_unlock();

		/* cleanup */
		g_free (widths);
		debug_exit ("gui_restore_column_widths");
	}
}


void
gui_save_column_widths (GtkWidget    * clist,
                        const char   * type)
{
	if (GTK_IS_CLIST (clist))
	{
		int i;
		int cols = GTK_CLIST (clist)->columns;
		GtkCList * list = GTK_CLIST (clist);
		char buf[1024];
		
		for (i=0; i<cols; i++)
		{
			g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_column_%d", type, i);
			pan_config_set_int (buf, (int)list->column[i].width);
		}


		pan_config_sync ();
	}
}

/**
***
**/

gboolean
gui_restore_window_size (GtkWidget     * widget,
                         const char    * key)
{
	int x, y, w, h;
	char buf[1024];
	int screen_w = gdk_screen_width ();
	int screen_h = gdk_screen_height ();
	const int fuzz = 10;
	gboolean retval = FALSE;

	/* sanity clause */
	g_return_val_if_fail (GTK_IS_WIDGET(widget), FALSE);
	g_return_val_if_fail (is_nonempty_string(key), FALSE);

	/* get the dimensions */
	g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_x", key);
	x = pan_config_get_int (buf);
	g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_y", key);
	y = pan_config_get_int (buf);
	g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_width", key);
	w = pan_config_get_int (buf);
	g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_height", key);
	h = pan_config_get_int (buf);

	/* move & resize */
	if ((0<x && x<screen_w-fuzz) && (0<y && y<screen_h-fuzz))
		gtk_widget_set_uposition (widget, x, y);
	if (w>0 && h>0) {
		gtk_window_set_default_size (GTK_WINDOW(widget), w, h);
		retval = TRUE;
	}

	return retval;
}
 
void
gui_save_window_size (GtkWidget     * widget,
                      const char    * key)
{
	char buf[1024];
	int x, y, w, h;

	/* sanity clause */
	g_return_if_fail (GTK_IS_WIDGET(widget));
	g_return_if_fail (widget->window!=NULL);
	g_return_if_fail (is_nonempty_string(key));

	gdk_window_get_position (widget->window, &x, &y);
	gdk_window_get_size (widget->window, &w, &h);

	g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_x", key);
	pan_config_set_int (buf, x);
	g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_y", key);
	pan_config_set_int (buf, y);
	g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_width", key);
	pan_config_set_int (buf, w);
	g_snprintf (buf, sizeof(buf), "/Pan/Geometry/%s_height", key);
	pan_config_set_int (buf, h);

	pan_config_sync ();


}

/*****
******
*****/

static int connection_size = -1;

static gboolean _is_connecting = FALSE;

static int
gui_refresh_connection_label_idle (gpointer data)
{
	char str[128];
	char tip[128];

	/* decide what to say */
	if (connection_size != 0) {
		g_snprintf (str, sizeof(str), _("%d @ %.1f KB/s"), connection_size, KBps);
		g_snprintf (tip, sizeof(tip), _("%d connections at %.1f KB per second"),
			connection_size, KBps);
	}
	else if (_is_connecting) {
		g_snprintf (str, sizeof(str), _("Connecting"));
		g_snprintf (tip, sizeof(tip), _("Connecting to a Remote News Server to Execute a Task"));

	}
	else {
		g_snprintf (str, sizeof(str), _("Offline"));
		g_snprintf (tip, sizeof(tip), _("No Connections"));
	}

	/* update the tip */
	pan_lock ();
	if (GTK_IS_LABEL(connection_qty_label)) {
		gtk_label_set_text (GTK_LABEL(connection_qty_label), str);
	if (GTK_IS_BUTTON(connection_qty_button))
		gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), connection_qty_button, tip, "");
	}
	pan_unlock ();

	return 0;
}

void
gui_set_connecting_flag (gboolean is_connecting)
{
	if (_is_connecting != is_connecting)
	{
		_is_connecting = is_connecting;
		gui_queue_add (gui_refresh_connection_label_idle, NULL);
	}
}

static void
gui_set_connection_size (int size)
{
	if (size != connection_size)
	{
		connection_size = size;
		log_add_va (LOG_INFO, _("News server connection count: %d"), size);
		gui_queue_add (gui_refresh_connection_label_idle, NULL);
	}
}

void
gui_inc_connection_size (int inc)
{
	gui_set_connection_size (connection_size + inc);
}

static int
gui_refresh_timer_cb (gpointer user_data)
{
	/* update KBps */
	if (1)
	{
		static GTimeVal last_time;
		static unsigned long last_KB = 0;
		static gboolean last_time_inited = FALSE;

		if (!last_time_inited)
		{
			g_get_current_time (&last_time);
			last_time_inited = TRUE;
		}
		else
		{
			double time_diff;
			double KB_diff;
			GTimeVal new_time;
			unsigned long new_KB;

			g_get_current_time (&new_time);
			time_diff = (double)(new_time.tv_sec - last_time.tv_sec) * G_USEC_PER_SEC;
			time_diff += (new_time.tv_usec - last_time.tv_usec);
			last_time = new_time;

			new_KB = pan_socket_get_total_xfer_K ();
			KB_diff = new_KB - last_KB;
			last_KB = new_KB;

			KB_diff *= (double)G_USEC_PER_SEC;
			KBps = KB_diff / time_diff;
		}
	}

	gui_queue_add (gui_refresh_connection_label_idle, NULL);

	return 1;
}

/*****
******
*****/

static int
gui_set_queue_size_idle (gpointer data)
{
	char str[128];
	char tip[128];
	ArgSet * argset = (ArgSet*) data;
	const guint running = GPOINTER_TO_UINT (argset_get (argset, 0));
	const guint size = GPOINTER_TO_UINT (argset_get (argset, 1));

	/* build the format strings */
	if (size == 0u) {
		g_snprintf (str, sizeof(str), _("No Tasks"));
		g_snprintf (tip, sizeof(tip), _("The Task Manager is Empty"));
	} else {
		g_snprintf (str, sizeof(str), _("Tasks: %u/%u"), running, size);
		g_snprintf (tip, sizeof(tip), _("%u Tasks Running, %u Tasks Total"), running, size);
	}

	/* update the gui */
	pan_lock();
	gtk_label_set_text (GTK_LABEL(queue_qty_label), str);
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), queue_qty_button, tip, "");
	pan_unlock();

	/* clean up */
	argset_free (argset);
	return 0;
}
void
gui_set_queue_size (guint running, guint size)
{
	gui_queue_add (gui_set_queue_size_idle,
	               argset_new2 (GUINT_TO_POINTER(running), GUINT_TO_POINTER(size)));
}

static int
gui_add_status_item_idle (gpointer p)
{
	int i;
	StatusItem * status;
	debug_enter ("gui_add_status_item");

	pan_lock ();

	for (i=0; i<VIEW_QTY; ++i)
		if (views[i].status_item == NULL)
			break;

       	status = STATUS_ITEM(p);
	if (i!=VIEW_QTY) {
		pan_object_ref (PAN_OBJECT(status)); /* unref in remove_status_item_idle */
		views[i].status_item = status;
		status_item_view_set_item_nolock (views[i].view, status);
	}

	pan_object_unref (PAN_OBJECT(status)); /* ref in item_active_changed_cb */
	pan_unlock ();
	debug_exit ("gui_add_status_item");
	return 0;
}

static int
gui_remove_status_item_idle (gpointer p)
{
	int i;
	StatusItem * status;
	debug_enter ("gui_remove_status_item");

	pan_lock ();

	status = STATUS_ITEM(p);
	for (i=0; i<VIEW_QTY; ++i)
		if (views[i].status_item == status)
			break;
	if (i!=VIEW_QTY) {
		GtkWidget * view = views[i].view;
		status_item_view_set_item_nolock (view, NULL);
		views[i].status_item = NULL;
		pan_object_unref (PAN_OBJECT(status)); /* ref in set_item_idle */
	}

	pan_unlock ();
	pan_object_unref (PAN_OBJECT(status)); /* ref in item_active_changed_cb */
	debug_exit ("gui_remove_status_item");
	return 0;
}

static void
turn_off_status_item_views_nolock (void)
{
	int i;
	debug_enter ("turn_off_status_item_views");

	for (i=0; i<VIEW_QTY; ++i) {
		StatusItem * item = views[i].status_item;
		GtkWidget * view = views[i].view;
		if (item != NULL) {
			status_item_view_set_item_nolock (view, NULL);
			views[i].status_item = NULL;
		}
	}

	debug_exit ("turn_off_status_item_views");
}

static void
status_item_active_changed_cb (gpointer call_obj, gpointer call_arg, gpointer user_data)
{
	StatusItem * item = STATUS_ITEM(call_obj);
	gboolean is_active = GPOINTER_TO_INT(call_arg) != 0;
	debug_enter ("status_item_active_changed_cb");

	/**
	***  Ref the item to make sure it's still alive when the
	***  idle func is run.  The idle func unrefs it to balance
	***  this out.
	**/

	pan_object_ref (PAN_OBJECT(item));

	/**
	***  Push the GUI work to the main thread
	**/

	gui_queue_add (is_active ? gui_add_status_item_idle : gui_remove_status_item_idle, item);


	debug_exit ("status_item_active_changed_cb");
}

/*****
******
*****/

void
pan_widget_set_font (GtkWidget * w, const char * font_name)
{
	PangoFontDescription * font_desc;

	g_return_if_fail (GTK_IS_WIDGET(w));
	g_return_if_fail (is_nonempty_string(font_name));

	font_desc = pango_font_description_from_string (font_name);
	gtk_widget_modify_font (w, font_desc);
	pango_font_description_free (font_desc);
}

/*****
******  This is for delegating small, async GUI updates to the main thread
******  to reduce the number of threads contending for a GUI lock.
*****/

static GStaticMutex gui_queue_mutex = G_STATIC_MUTEX_INIT;

static GSList * gui_queue_slist = NULL;

typedef struct
{
	GSourceFunc func;
	gpointer user_data;
}
GuiQueueItem;

static int
gui_queue_timer_cb (gpointer user_data)
{
	GSList * list;

	/* get the work list */
	g_static_mutex_lock (&gui_queue_mutex);
	list = gui_queue_slist;
	gui_queue_slist = NULL;
	g_static_mutex_unlock (&gui_queue_mutex);

	if (list != NULL)
	{
		GSList * l;

		/* gui_queue_slist is LIFO for efficiency of appending nodes;
		 * reverse the nodes to make list FIFO */
		list = g_slist_reverse (list);

		/* run the work functions */
		for (l=list; l!=NULL; l=l->next)
		{
			GuiQueueItem * queue_item = (GuiQueueItem*) l->data;
			(*queue_item->func)(queue_item->user_data);
			g_free (queue_item);
		}
		g_slist_free (list);
	}

	return 1;
}

void
gui_queue_add (GSourceFunc func, gpointer user_data)
{
	GuiQueueItem * item;

	/* create the new queue item */	
	item = g_new (GuiQueueItem, 1);
	item->func = func;
	item->user_data = user_data;

	/* add it to the queue */
	g_static_mutex_lock (&gui_queue_mutex);
	gui_queue_slist = g_slist_prepend (gui_queue_slist, item);
	g_static_mutex_unlock (&gui_queue_mutex);
}

/***
****
***/

static GtkItemFactoryEntry main_menu_entries[MAIN_MENU_ENTRIES_QTY] =
{
	/*   */ {N_("/_File"), NULL, NULL, 0, "<Branch>"},
	/*   */ {N_("/_File/---"), NULL, NULL, 0, "<Tearoff>"},
	/* s */ {N_("/_File/_Save Article As..."), "<control>S", article_action_selected_save_as, 0, "<StockItem>", GTK_STOCK_SAVE_AS},
	/* t */ {N_("/_File/Save Article A_ttachments"), "<shift>S", article_action_selected_save_attachments, 0, "<StockItem>", GTK_STOCK_SAVE},
	/*   */ {N_("/_File/---"), NULL, NULL, 0, "<Separator>"},
	/* p */ {N_("/_File/_Print Article..."), "<Control>P", print_cb, 0, "<StockItem>", GTK_STOCK_PRINT},
	/*   */ {N_("/_File/---"), NULL, NULL, 0, "<Separator>"},
	/* n */ {N_("/_File/_New Folder..."), NULL, group_new_folder_dialog, 0, NULL, NULL},
	/* d */ {N_("/_File/_Delete Folder..."), NULL, group_action_selected_destroy, 0, "<StockItem>", GTK_STOCK_DELETE},
	/* e */ {N_("/_File/_Edit Article in Folder"), NULL, article_action_edit_selected, 0, NULL},
	/*   */ {N_("/_File/---"), NULL, NULL, 0, "<Separator>"},
	/* x */ {N_("/_File/E_xit"), "<control>Q", exit_cb, 0, "<StockItem>", GTK_STOCK_QUIT},

	/*   */ {N_("/_Edit"), NULL, NULL, 0, "<Branch>"},
	/*   */ {N_("/_Edit/---"), NULL, NULL, 0, "<Tearoff>"},
	/* g */ {N_("/_Edit/Select all _Groups"), "B", grouplist_select_all, 0, NULL},
	/* u */ {N_("/_Edit/Add S_uscribed to Selection"), "<control>B", grouplist_add_subscribed_to_selection_nolock, 0, NULL},
	/* r */ {N_("/_Edit/Deselect all G_roups"), "<shift><control>B", grouplist_deselect_all, 0, NULL},
	/*   */ {N_("/_Edit/---"), NULL, NULL, 0, "<Separator>"},
	/* a */ {N_("/_Edit/Select all _Articles"), "D", articlelist_select_all_nolock, 0, NULL},
	/* s */ {N_("/_Edit/Add _Subthreads to Selection"), "<control>D", articlelist_add_replies_to_selection_nolock, 0, NULL},
	/* t */ {N_("/_Edit/Add _Threads to Selection"), "<shift>D", articlelist_add_thread_to_selection_nolock, 0, NULL},
	/* l */ {N_("/_Edit/Deselect a_ll Articles"), "<shift><control>D", articlelist_deselect_all_nolock, 0, NULL},
	/*   */ {N_("/_Edit/---"), NULL, NULL, 0, "<Separator>"},
	/* b */ {N_("/_Edit/Select Article _Body"), NULL, text_select_all, 0, NULL},
	/*   */ {N_("/_Edit/---"), NULL, NULL, 0, "<Separator>"},
	/* f */ {N_("/_Edit/_Find..."), "<control>F", articlelist_find_text_cb, 0, "<StockItem>", GTK_STOCK_FIND},
	/* n */ {N_("/_Edit/Find _Next..."), "<control><shift>F", articlelist_find_next_cb, 0, "<StockItem>", GTK_STOCK_FIND},
	/*   */ {N_("/_Edit/---"), NULL, NULL, 0, "<Separator>"},
	/* p */ {N_("/_Edit/_Preferences..."), NULL, prefs_spawn, 0, "<StockItem>", GTK_STOCK_PREFERENCES},

	{N_("/_View"), NULL, NULL, 0, "<Branch>"},
	{N_("/_View/---"), NULL, NULL, 0, "<Tearoff>"},
	{N_("/_View/_Current Filter..."), "<shift>F", article_toolbar_edit_filter, 0, "<ImageItem>", icon_filter},
	{N_("/_View/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_View/View Pa_nes as Tabs"), "Z", zoom_cb, 0, "<StockItem>", GTK_STOCK_JUMP_TO},
	{N_("/_View/_View Group Tab"), "1", zoom_page_cb, GROUPS_PAGE, "<StockItem>", GTK_STOCK_JUMP_TO},
	{N_("/_View/View Header Ta_b"), "2", zoom_page_cb, ARTICLELIST_PAGE, "<StockItem>", GTK_STOCK_JUMP_TO},
	{N_("/_View/View _Article Tab"), "3", zoom_page_cb, MESSAGE_PAGE, "<StockItem>", GTK_STOCK_JUMP_TO},
	{N_("/_View/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_View/Sho_w Group Pane"), "<control>1", toggle_pane_visibility, ARTICLELIST_PAGE, "<ToggleItem>"},
	{N_("/_View/Show Hea_der Pane"), "<control>2", toggle_pane_visibility, ARTICLELIST_PAGE, "<ToggleItem>"},
	{N_("/_View/Show Art_icle Pane"), "<control>3", toggle_pane_visibility, MESSAGE_PAGE, "<ToggleItem>"},
	{N_("/_View/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_View/Collapse Names in Gro_up Pane"), NULL, collapse_group_names_cb, 0, "<ToggleItem>"},
	{N_("/_View/Thr_ead Header Pane"), "<control>E", thread_articles_cb, 0, "<ToggleItem>"},
	{N_("/_View/_Rot13 Selected Text"), "<control>R", rot13_cb, 0, NULL},
	{N_("/_View/Fi_ll & Wrap Article Body"), "W", fill_body_cb, 0, "<ToggleItem>"},
	{N_("/_View/Mute _Quoted Text"), "Q", mute_quoted_text_cb, 0, "<ToggleItem>"},
	{N_("/_View/Show All _Headers in Article Pane"), "H", show_all_headers_cb, 0, "<ToggleItem>"},
	{N_("/_View/Use _Monospace Font in Article Pane"), "C", use_fixed_font_cb, 0, "<ToggleItem>"},
	{N_("/_View/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_View/E_xpand Selected Threads"), NULL, articlelist_expand_selected_threads, 0, NULL},
	{N_("/_View/Expand all Thread_s"), NULL, articlelist_expand_all_threads, 0, NULL},
	{N_("/_View/C_ollapse Selected Threads"), NULL, articlelist_collapse_selected_threads, 0, NULL},
	{N_("/_View/Colla_pse all Thread_s"), NULL, articlelist_collapse_all_threads, 0, NULL},

	{N_("/_Servers"), NULL, NULL, 0, "<Branch>"}, /* <-- the server list gets inserted here */
	{N_("/_Servers/---"), NULL, NULL, 0, "<Tearoff>"},
	{N_("/_Servers/Select _Server"),  NULL, NULL, 0, "<Branch>"},
	{N_("/_Servers/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_Servers/Get List of _All Groups..."), NULL, grouplist_get_all, 0, NULL},
	{N_("/_Servers/Get List of _New Groups..."), NULL, grouplist_get_new, 0, NULL},
	{N_("/_Servers/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_Servers/_Import .newsrc..."), NULL, server_import_newsrc_dialog_menu, 0, "<StockItem>", GTK_STOCK_REDO},
	{N_("/_Servers/_Export .newsrc..."), NULL, server_export_newsrc_dialog_menu, 0, "<StockItem>", GTK_STOCK_UNDO},

	/*   */ {N_("/_Newsgroups"), NULL, NULL, 0, "<Branch>"},
	/*   */ {N_("/_Newsgroups/---"), NULL, NULL, 0, "<Tearoff>"},
	/* r */ {N_("/_Newsgroups/Mark Group _Read"), "<control><shift>M", group_action_selected_mark_read, 0, "<StockItem>", GTK_STOCK_CLEAR},
	/* d */ {N_("/_Newsgroups/_Delete Group's Articles"), "<shift><control>Delete", group_action_selected_empty, 0, "<StockItem>", GTK_STOCK_DELETE},
	/*   */ {N_("/_Newsgroups/---"), NULL, NULL, 0, "<Separator>"},
	/* h */ {N_("/_Newsgroups/Get New _Headers"), "<control>A", group_action_selected_download_new, 0, NULL},
	/* b */ {N_("/_Newsgroups/Get New Headers and _Bodies"), "<shift><control>A", group_action_selected_download_new_and_bodies, 0, NULL},
	/* o */ {N_("/_Newsgroups/More Download _Options..."), "<control>O", group_action_selected_download_dialog, 0, NULL},
	/* c */ {N_("/_Newsgroups/Refresh Article _Counts"), NULL, group_action_selected_update_count_info, 0, "<StockItem>", GTK_STOCK_REFRESH},
	/*   */ {N_("/_Newsgroups/---"), NULL, NULL, 0, "<Separator>"},
	/* s */ {N_("/_Newsgroups/_Subscribe"), NULL, group_action_selected_subscribe, 0, "<ImageItem>", icon_stock_book_red},
	/* u */ {N_("/_Newsgroups/_Unsubscribe"), NULL, group_action_selected_unsubscribe, 0, NULL},
	/* p */ {N_("/_Newsgroups/Group _Properties..."), NULL, group_action_selected_properties, 0, "<StockItem>", GTK_STOCK_PROPERTIES},
	/*   */ {N_("/_Newsgroups/---"), NULL, NULL, 0, "<Separator>"},
	/* t */ {N_("/_Newsgroups/Dele_te Group"), NULL, group_action_selected_destroy, 0, "<StockItem>", GTK_STOCK_DELETE},

	/*   */ {N_("/_Articles"), NULL, NULL, 0, "<Branch>"},
	/*   */ {N_("/_Articles/---"), NULL, NULL, 0, "<Tearoff>"},
	/* r */ {N_("/_Articles/Mark _Read"), "M", articlelist_selected_mark_read_nolock, 0, NULL},
	/* u */ {N_("/_Articles/Mark _Unread"), NULL, articlelist_selected_mark_unread_nolock, 0, NULL},
	/*   */ {N_("/_Articles/---"), NULL, NULL, 0, "<Separator>"},
	/* o */ {N_("/_Articles/D_ownload Flagged"), "<control><shift>J", articlelist_download_flagged_bodies, 0, "<ImageItem>", icon_blue_flag},
	/* f */ {N_("/_Articles/_Flag"), "J", articlelist_selected_flag_for_dl_nolock, 0, "<ImageItem>", icon_blue_flag},
	/* n */ {N_("/_Articles/U_nflag"), "<control>J", articlelist_selected_unflag_for_dl_nolock, 0, NULL},
	/*   */ {N_("/_Articles/---"), NULL, NULL, 0, "<Separator>"},
	/* l */ {N_("/_Articles/Down_load"), NULL, articlelist_selected_download_nolock, 0, NULL},
	/*   */ {N_("/_Articles/---"), NULL, NULL, 0, "<Separator>"},
	/* w */ {N_("/_Articles/_Watch Thread"), "<control>W", articlelist_selected_thread_watch_nolock, 0, "<ImageItem>", icon_watched},
	/* i */ {N_("/_Articles/_Ignore Thread"), "<control>I", articlelist_selected_thread_ignore_nolock, 0, "<StockItem>", GTK_STOCK_CLEAR},
	/* z */ {N_("/_Articles/Add to Bo_zo or Spam Filter..."), "<control>Z", articlelist_spamize_nolock, 0, "<ImageItem>", icon_bozo},
	/*   */ {N_("/_Articles/---"), NULL, NULL, 0, "<Separator>"},
	/* p */ {N_("/_Articles/Cop_y to Folder"), NULL, articlelist_selected_copy_to_folder_nolock, 0, "<StockItem>", GTK_STOCK_COPY},
	/*   */ {N_("/_Articles/---"), NULL, NULL, 0, "<Separator>"},
	/* c */ {N_("/_Articles/_Cancel..."), NULL, articlelist_selected_cancel, 0, "<StockItem>", GTK_STOCK_DELETE},
	/* s */ {N_("/_Articles/_Supersede..."), NULL, articlelist_selected_supersede, 0, "<StockItem>", GTK_STOCK_CONVERT},
	/*   */ {N_("/_Articles/---"), NULL, NULL, 0, "<Separator>"},
	/* d */ {N_("/_Articles/_Delete"), "Delete", article_action_delete_selected_articles, 0, "<StockItem>", GTK_STOCK_DELETE},

	{N_("/_Go"), NULL, NULL, 0, "<Branch>"},
	{N_("/_Go/---"), NULL, NULL, 0, "<Tearoff>"},
	{N_("/_Go/_Load Articles When Selected:"), NULL, navigate_read_on_select_cb, 0, "<ToggleItem>"},
	{N_("/_Go/_Next Unread _Article"), "N", articlelist_read_next_unread, 0, "<StockItem>", GTK_STOCK_GO_FORWARD},
	{N_("/_Go/Ne_xt Article"), "<control>N", gui_read_next, 0, "<StockItem>", GTK_STOCK_GO_FORWARD},
	{N_("/_Go/Next Unread _Thread"), "T", articlelist_read_next_unread_thread, 0, "<StockItem>", GTK_STOCK_GO_FORWARD},
	{N_("/_Go/Next Threa_d"), "<control>T", articlelist_read_next_thread, 0, "<StockItem>", GTK_STOCK_GO_FORWARD},
	{N_("/_Go/Pre_vious Article"), "V", articlelist_read_prev, 0, "<StockItem>", GTK_STOCK_GO_BACK},
	{N_("/_Go/Previo_us Thread"), "<control>V", articlelist_read_prev_thread, 0, "<StockItem>", GTK_STOCK_GO_BACK},
	{N_("/_Go/Previous Article _Read"), "<shift>V", articlelist_read_prev_read, 0, "<StockItem>", GTK_STOCK_GO_BACK},
	{N_("/_Go/_Parent Article"), "U", articlelist_read_parent, 0, "<StockItem>", GTK_STOCK_GO_UP},
	{N_("/_Go/T_op of Thread"), "<control>U", articlelist_read_top_of_thread, 0, "<StockItem>", GTK_STOCK_GOTO_TOP},
	{N_("/_Go/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_Go/Next Unread _Group"), "G", grouplist_activate_next_unread_group, 0, "<StockItem>", GTK_STOCK_GO_FORWARD},
	{N_("/_Go/N_ext Group"), "<control>G", grouplist_activate_next_group, 0, "<StockItem>", GTK_STOCK_GO_FORWARD},

	{N_("/_Post"), NULL, NULL, 0, "<Branch>"},
	{N_("/_Post/---"), NULL, NULL, 0, "<Tearoff>"},
	{N_("/_Post/_Post to Newsgroup"), "P", message_post_window, 0, "<ImageItem>", icon_stock_mail_compose},
	{N_("/_Post/_Follow-Up to Newsgroup"), "F", message_followup_window, 0, "<ImageItem>", icon_stock_mail_reply},
	{N_("/_Post/_Reply by E-Mail"), "R", message_reply_window, 0, "<ImageItem>", icon_stock_mail_reply},
	{N_("/_Post/Follow-Up _and Reply"), NULL, message_followup_reply_window, 0, "<ImageItem>", icon_stock_mail_reply},
	{N_("/_Post/For_ward Article by E-Mail"), NULL, message_forward_window, 0, "<ImageItem>", icon_stock_mail_forward},
	{N_("/_Post/Send Pending _Messages"), NULL, flush_sendlater_articles, 0, "<ImageItem>", icon_stock_mail_send},

	{N_("/_Tools"), NULL, NULL, 0, "<Branch>"},
	{N_("/_Tools/---"), NULL, NULL, 0, "<Tearoff>"},
	{N_("/_Tools/_Posting Profiles..."), NULL, identity_cb, 0, "<ImageItem>", icon_by_me},
	{N_("/_Tools/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_Tools/_Custom Filters..."), NULL, filter_custom_cb, 0, "<ImageItem>", icon_filter},
	{N_("/_Tools/_Rules..."), NULL, rules_cb, 0, NULL},
	{N_("/_Tools/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_Tools/_Online Settings..."), NULL, prefs_spawn_to_news_connections, 0, "<ImageItem>", icon_server},
	{N_("/_Tools/_Task Manager..."), NULL, task_manager_spawn, 0, NULL},
	{N_("/_Tools/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_Tools/_Log Viewer..."), NULL, log_button_clicked, 0, "<ImageItem>", GTK_STOCK_DIALOG_INFO},

	{N_("/_Help"), NULL, NULL, 0, "<Branch>"},
	{N_("/_Help/---"), NULL, NULL, 0, "<Tearoff>"},
	{N_("/_Help/Pan _Home Page..."), NULL, pan_homepage_url, 0, "<StockItem>", GTK_STOCK_HOME},
	{N_("/_Help/Online Users` _Manual..."), NULL, pan_manual_url, 0, "<ImageItem>", icon_stock_book_red},
	{N_("/_Help/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_Help/_Feedback..."), NULL, send_feedback_cb, 0, "<ImageItem>", icon_stock_mail_compose},
	{N_("/_Help/Report a _Bug..."), NULL, pan_bug_url, 0, "<StockItem>", GTK_STOCK_DIALOG_ERROR},
	{N_("/_Help/---"), NULL, NULL, 0, "<Separator>"},
	{N_("/_Help/_About..."), NULL, dialog_about_cb, 0, NULL}
};
