/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Author: Charles Kerr <charles@rebelbase.com>
 *
 * Copyright (C) 2000, 2001  Pan Development Team <pan@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <pan/base/debug.h>
#include <pan/base/pan-i18n.h>
#include <pan/base/pan-glib-extensions.h>

#include <pan/gui.h>
#include <pan/prefs.h>
#include <pan/queue.h>
#include <pan/task-manager.h>
#include <pan/util.h>

#include <pan/xpm/pan-pixbufs.h>

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

typedef struct
{
	GtkWidget    * window;
	GtkWidget    * top;
	GtkWidget    * list;

	guint          timeout_id;
	guint          moved_signal_id;

	gboolean       titlebar_needs_refresh;
	gboolean       dampen_move_feedback_loop;
}
ManagerUI;

/*********************
**********************  Variables
*********************/

extern GtkTooltips * ttips;

static ManagerUI * manager_ui = NULL;

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PRIVATE ROUTINES
************/

/**
***  Internal Utility
**/

static GPtrArray*
get_selected_tasks (ManagerUI * ui)
{
	GPtrArray * a = g_ptr_array_new ();
	GtkCList *clist = GTK_CLIST(ui->list);
        GList * l;
	for (l=clist->selection; l!=NULL; l=l->next)
	{
		const int index = GPOINTER_TO_INT (l->data);
		Task * const task = gtk_clist_get_row_data (clist, index);
		g_ptr_array_add (a, task);
	}
	return a;
}

/**
*** Helper functions for user interactions
**/

static void
move_selected_tasks_to_top (ManagerUI * ui)
{
	GtkCList * clist;
	GList * l;
	gint new_row;
	debug_enter ("move_selected_tasks_to_top");

	/* sanity clause */
	g_return_if_fail (ui!=NULL);

	clist = GTK_CLIST(ui->list);
	new_row = -1;
	for (l=clist->selection; l!=NULL; l=l->next)
	{
		const int old_row = GPOINTER_TO_INT (l->data);
		Task * const task = gtk_clist_get_row_data (clist, old_row);
		queue_task_move (task, ++new_row);
	}

	debug_exit ("move_selected_tasks_to_top");
}

static void
move_selected_tasks_to_bottom (ManagerUI * ui)
{
	GtkCList * clist;
	GList * l;
	gint new_row;
	debug_enter ("move_selected_tasks_to_bottom");

	/* sanity clause */
	g_return_if_fail (ui!=NULL);

	clist = GTK_CLIST(ui->list);
	new_row = 0;
	for (l=clist->selection; l!=NULL; l=l->next)
	{
		const int old_row = GPOINTER_TO_INT (l->data);
		Task * const task = gtk_clist_get_row_data (clist, old_row);
		queue_task_move (task, clist->rows - ++new_row);
	}

	debug_exit ("move_selected_tasks_to_bottom");
}

static void
cancel_selected_tasks (ManagerUI * ui)
{
	GPtrArray * a;
        debug_enter ("cancel_selected_tasks");

	/* sanity clause */
	g_return_if_fail (ui!=NULL);

	a = get_selected_tasks (ui);
	pan_g_ptr_array_foreach (a, (GFunc)queue_task_remove, NULL);
	g_ptr_array_free (a, TRUE);

        debug_exit ("cancel_selected_tasks");
}

static void
select_task (ManagerUI * ui, gboolean first)
{
	GtkCList * clist;
	gint       row;
	debug_enter ("select_task");

	/* sanity clause */
	g_return_if_fail (ui!=NULL);

	clist = GTK_CLIST(ui->list);
	row = first ? 0 : clist->rows-1;
	
	gtk_clist_freeze(clist);
	gtk_clist_unselect_all (clist);
	gtk_clist_select_row (clist, row, 0);
	clist->focus_row = row;
	gtk_clist_thaw(clist);

	debug_exit("select_task");
}


/**
***  User Interactions
**/

static void
pause_toggled_cb (GtkToggleButton  * tb,
                  gpointer           user_data)
{
	gboolean paused = gtk_toggle_button_get_active (tb);

	if (paused != queue_is_paused())
		queue_set_paused (paused);

	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), GTK_WIDGET(tb), paused
		? _("The Task Manager is Paused")
		: _("The Task Manager is not Paused"), "");
}

static void
up_clicked_cb (GtkButton   * button,
               gpointer      user_data)
{
        ManagerUI * ui;
	GtkCList * clist;
	GList * l;
	gint last_index;
	debug_enter ("up_clicked_cb");

        ui = (ManagerUI*) user_data;
	clist = GTK_CLIST(ui->list);
	last_index = -1;
	for (l=clist->selection; l!=NULL; l=l->next)
	{
		const int index = GPOINTER_TO_INT (l->data);
		Task * const task = gtk_clist_get_row_data (clist, index);
		int new_index = MAX (last_index+1, index-1);
		queue_task_move (task, new_index);
		last_index = new_index;
	}

	debug_exit ("up_clicked_cb");
}

static void
top_clicked_cb (GtkButton    * button,
                gpointer       user_data)
{
	debug_enter ("top_clicked_cb");

	move_selected_tasks_to_top ((ManagerUI *) user_data);

	debug_exit ("top_clicked_cb");
}

static void
down_clicked_cb (GtkButton   * button,
                 gpointer      user_data)
{
	ManagerUI * ui;
	GtkCList * clist;
	GList * l;
	debug_enter ("down_clicked_cb");

	ui = (ManagerUI*) user_data;
	clist = GTK_CLIST(ui->list);
	for (l=g_list_last(clist->selection); l!=NULL; l=l->prev)
	{
		const int index = GPOINTER_TO_INT (l->data);
		Task * const task = gtk_clist_get_row_data (clist, index);
		queue_task_move (task, index+1);
	}

	debug_exit ("down_clicked_cb");
}

static void
bottom_clicked_cb (GtkButton   * button,
                   gpointer      user_data)
{
	debug_enter ("bottom_clicked_cb");

	move_selected_tasks_to_bottom ((ManagerUI *) user_data);

	debug_exit ("bottom_clicked_cb");
}

static void
cancel_clicked_cb (GtkButton   * button,
                   gpointer      user_data)
{
	debug_enter ("cancel_clicked");

	cancel_selected_tasks ((ManagerUI *) user_data);

	debug_exit ("cancel_clicked");
}

static void
clear_queue_clicked_cb (GtkButton   * b,
                        gpointer      user_data)
{
	guint i;
	GPtrArray * tasks;
	debug_enter ("clear_queue_clicked_cb");

	tasks = queue_get_tasks ();
	for (i=0; i<tasks->len; ++i) {
		Task * task = TASK(g_ptr_array_index (tasks,i));
		if (queue_get_task_status(task)==QUEUE_TASK_STATUS_QUEUED)
			queue_task_remove(task);
	}

	/* cleanup */
	pan_g_ptr_array_foreach (tasks, (GFunc)pan_object_unref, NULL);
	g_ptr_array_free (tasks, TRUE);

	debug_exit ("clear_queue_clicked_cb");
}

static void
requeue_cb (GtkButton   * button,
            gpointer      user_data)
{
	guint i;
	ManagerUI * ui;
	GPtrArray * a;
	debug_enter ("requeue_cb");

	ui = (ManagerUI*) user_data;
	a = get_selected_tasks (ui);
	for (i=0; i!=a->len; ++i) {
		Task * task = (Task*) g_ptr_array_index (a, i);
		if (queue_get_task_status(task) == QUEUE_TASK_STATUS_FAILED)
			queue_task_requeue_failed (task, 0);
	}
	g_ptr_array_free (a, TRUE);

	debug_exit ("requeue_cb");
}


/**
***  Display
**/

static int
get_secs_left (const Task * task)
{
	int retval = 0;

	if (task->sock)
	{
		double elapsed_time = time(0) - task->sock->byte_count_start_time;
		double progress = status_item_get_progress_of_100 (STATUS_ITEM(task)) / 100.0;
		double safe_progress = MAX(progress, 0.01); /* don't divide by 0 */
		const double estimated_total_time = elapsed_time/safe_progress;
		retval = estimated_total_time - elapsed_time;
	}

	return retval;
}

static void
get_percent_str (const Task   * task,
                 gchar        * buf,
                 gint           buf_size)
{
	int percent = 0;

	if (task != NULL)
		percent = status_item_get_progress_of_100 (STATUS_ITEM(task));

	if (percent == 0)
		*buf = '\0';
	else
		g_snprintf (buf, buf_size, "%d%%", percent);
}

static void
get_xfer_str (const Task    * task,
              gchar         * buf,
              gint            buf_size)
{
	if (!task->sock)
	{
		*buf = '\0';
	}
	else
	{
		int secs_left = get_secs_left (task);
		const int h = secs_left / 3600;
		const int m = (secs_left % 3600) / 60;
		const int s = secs_left % 60;
		const double KBps = pan_socket_get_xfer_rate_KBps (task->sock);
		g_snprintf (buf, buf_size, _("%.2f KB/s, %d:%02d:%02d remaining"), KBps, h, m, s);
	}
}

static void
get_status_str (const Task     * task,
                gchar          * buf,
                gint             buf_size)
{
	const char* status_str;
	QueueTaskStatus status = queue_get_task_status(task);

	switch (status)
	{
		case QUEUE_TASK_STATUS_NOT_QUEUED:
			status_str = _("Not Queued");
			break;
		case QUEUE_TASK_STATUS_QUEUED:
			status_str = _("Queued");
			break;
		case QUEUE_TASK_STATUS_RUNNING:
			status_str = _("Running");
			break;
		case QUEUE_TASK_STATUS_FAILED:
			status_str = _("Failed");
			break;
		case QUEUE_TASK_STATUS_ABORTING:
			status_str = _("Aborting");
			break;
		case QUEUE_TASK_STATUS_CONNECTING:
			status_str = _("Connecting");
			break;
		default:
			status_str = _("???");
			break;
	}

	if (!task->server->online && status==QUEUE_TASK_STATUS_QUEUED)
		g_snprintf (buf, buf_size, _("Offline"));
	else if (!task->tries)
		g_snprintf (buf, buf_size, "%s", status_str);
	else
		g_snprintf (buf, buf_size, _("%s (%d tries)"), status_str, task->tries);
}


static void
add_task_to_list (ManagerUI     * ui,
                  Task          * task,
                  int             index)
{
	int row;
	char statusbuf[64];
	char percentbuf[8];
	char xferbuf[64];
	GtkCList * list = GTK_CLIST(ui->list);
	char * description = status_item_describe(STATUS_ITEM(task));
	char * text [5];
	char * freeme[5] = { NULL, NULL, NULL, NULL, NULL };

	get_xfer_str (task, xferbuf, sizeof(xferbuf));
	get_status_str (task, statusbuf, sizeof(statusbuf));
	get_percent_str (task, percentbuf, sizeof(percentbuf));

	text[0] = (char*) pan_utf8ize (statusbuf, -1, freeme+0);
	text[1] = (char*) pan_utf8ize (percentbuf, -1, freeme+1);
	text[2] = (char*) pan_utf8ize (task->server->name, -1, freeme+2);
	text[3] = (char*) pan_utf8ize (xferbuf, -1, freeme+3);
	text[4] = (char*) pan_utf8ize (description, -1, freeme+4);

	pan_lock ();
	row = gtk_clist_insert (list, index, text);
	gtk_clist_set_row_data (list, row, task);
	pan_unlock ();

	g_free (description);

	/* cleanup */
	for (row=0; row<G_N_ELEMENTS(freeme); ++row)
		if (freeme[row] != NULL)
			g_free (freeme[row]);
}


static void
ui_populate (ManagerUI *ui)
{
	guint i;
	GPtrArray * tasks;

	tasks = queue_get_tasks ();
	for (i=0; i<tasks->len; ++i)
		add_task_to_list (ui, TASK(g_ptr_array_index(tasks,i)), i);

	/* cleanup */
	pan_g_ptr_array_foreach (tasks, (GFunc)pan_object_unref, NULL);
	g_ptr_array_free (tasks, TRUE);
}

static void
queue_row_move_cb (GtkCList   * clist,
                   int          old_index,
                   int          new_index,  
                   gpointer     user_data)
{
	ManagerUI * ui;
	Task * task;
	debug_enter ("queue_row_move_cb");

	ui = (ManagerUI*) user_data;
	task = TASK(gtk_clist_get_row_data (clist, old_index));
	if (!ui->dampen_move_feedback_loop)
		queue_task_move (task, new_index);

	debug_exit ("queue_row_move_cb");
}


/**
***  Queue Changes
**/

static void
task_added_cb (gpointer call_obj, gpointer call_arg, gpointer client_arg)
{
	Task * task = TASK(call_obj);
	int index = GPOINTER_TO_INT(call_arg);
	ManagerUI * ui = (ManagerUI*) client_arg;

	add_task_to_list (ui, task, index);
}

static void
task_removed_cb (gpointer call_obj, gpointer call_arg, gpointer client_arg)
{
	Task * task = TASK(call_obj);
	int index = GPOINTER_TO_INT(call_arg);
	ManagerUI * ui = (ManagerUI*) client_arg;
	GtkCList * list = GTK_CLIST(ui->list);

	g_return_if_fail (gtk_clist_get_row_data(list,index)==task);
	pan_lock ();
	gtk_clist_remove (list, index);
	pan_unlock ();
}

static void
task_moved_cb (gpointer call_obj, gpointer call_arg, gpointer client_arg)
{
	ManagerUI * ui = (ManagerUI*) client_arg;
	int old_row = ((int*)call_arg)[0];
	int new_row = ((int*)call_arg)[1];

	pan_lock ();
	gtk_signal_disconnect (GTK_OBJECT(ui->list), ui->moved_signal_id);
	gtk_clist_row_move (GTK_CLIST(ui->list), old_row, new_row);
	ui->moved_signal_id = g_signal_connect (ui->list, "row_move", G_CALLBACK(queue_row_move_cb), ui);
	pan_unlock ();
}

static gboolean
gui_key_press_cb (GtkWidget      * widget,
                  GdkEventKey    * event,
                  gpointer         data)
{
	ManagerUI * ui;
	GtkCList  * clist;
	debug_enter ("gui_key_press_cb");

	ui    = (ManagerUI*) data;
	clist = GTK_CLIST (ui->list);

        switch (event->keyval)
        {
                case GDK_Delete:
			cancel_selected_tasks (ui);
			break;
                case GDK_Home:
                        if (event->state & GDK_SHIFT_MASK)
			{
				move_selected_tasks_to_top (ui);
			}
			else
			{
				select_task (ui, TRUE);
			}
			break;
                case GDK_End:
                        if (event->state & GDK_SHIFT_MASK)
			{
				move_selected_tasks_to_bottom (ui);
			}
			else
			{
				select_task (ui, FALSE);
			}
			break;
                default:
			break;
        }

	debug_exit ("gui_key_press_cb");
        return TRUE;
}

static void
update_titlebar_lock (ManagerUI * ui)
{
	int running, queued, failed;
	guint i;
	GPtrArray * tasks;
	gchar* pch;

	g_return_if_fail (ui!=NULL);

	/* get information about the tasks */
	running = queued = failed = 0;
	tasks = queue_get_tasks ();
	for (i=0; i<tasks->len; ++i) {
		Task * t = TASK(g_ptr_array_index(tasks,i));
		QueueTaskStatus status = queue_get_task_status (t);
	
		if (status==QUEUE_TASK_STATUS_QUEUED)
			++queued;
		else if (status==QUEUE_TASK_STATUS_RUNNING)
			++running;
		else if (status==QUEUE_TASK_STATUS_ABORTING || status==QUEUE_TASK_STATUS_FAILED)
			++failed;
	}

	/* create a titlebar */
	if (failed)
		pch = g_strdup_printf (_("Pan %s Task Manager (%d Queued, %d Running, %d Failed)"), VERSION, queued, running, failed);
	else if (queued || running)
		pch = g_strdup_printf (_("Pan %s Task Manager (%d Queued, %d Running)"), VERSION, queued, running);
	else 
		pch = g_strdup_printf (_("Pan %s Task Manager"), VERSION);

	/* show the titlebar */
	pan_lock ();
	gtk_window_set_title (GTK_WINDOW(ui->window), pch);
	pan_unlock ();
	g_free (pch);

	/* cleanup */
	pan_g_ptr_array_foreach (tasks, (GFunc)pan_object_unref, NULL);
	g_ptr_array_free (tasks, TRUE);
}

static int
periodic_update_timer (gpointer user_data)
{
	ManagerUI * ui = (ManagerUI*) user_data;
	GtkCList * clist = GTK_CLIST(ui->list);
	gint row;

	debug0 (DEBUG_QUEUE, "task manager periodic ui refresh");

	/* update the ui */
	pan_lock ();
	gtk_clist_freeze (clist);
	for (row=0; row<clist->rows; ++row)
	{
		if (gtk_clist_row_is_visible (clist, row))
		{
			char buf[256];
			char * pch;
			char * freeme = NULL;
			Task * task;

			task = gtk_clist_get_row_data(clist, row);
			pan_object_ref (PAN_OBJECT(task));

			get_status_str (task, buf, sizeof(buf));
			gtk_clist_set_text (clist, row, 0, pan_utf8ize (buf, -1, &freeme));
			g_free (freeme);

			get_percent_str (task, buf, sizeof(buf));
			gtk_clist_set_text (clist, row, 1, pan_utf8ize (buf, -1, &freeme));
			g_free (freeme);

			gtk_clist_set_text (clist, row, 2, pan_utf8ize (task->server->name, -1, &freeme));
			g_free (freeme);

			get_xfer_str (task, buf, sizeof(buf));
			gtk_clist_set_text (clist, row, 3, pan_utf8ize (buf, -1, &freeme));
			g_free (freeme);

			pch = status_item_describe(STATUS_ITEM(task));
			gtk_clist_set_text (clist, row, 4, pan_utf8ize (pch, -1, &freeme));
			g_free (freeme);
			g_free (pch);

			pan_object_unref (PAN_OBJECT(task));
		}
	}

	gtk_clist_thaw (clist);
	pan_unlock ();

	if (ui->titlebar_needs_refresh)
	{
		update_titlebar_lock (ui);
		ui->titlebar_needs_refresh = FALSE;
	}

	return 1;
}

static gint
window_delete_event_cb (GtkWidget * w, GdkEvent * e, gpointer data)
{
	ManagerUI * ui = (ManagerUI*) data;
	gui_save_window_size (ui->window, "task_manager_2");
	gui_save_column_widths (ui->list, "task_manager_2");
	return FALSE;
} 
static void
ui_destroy_cb (GtkObject * o, gpointer user_data)
{
	ManagerUI *ui = (ManagerUI*) user_data;

	pan_callback_remove (queue_task_added, task_added_cb, ui);
	pan_callback_remove (queue_task_removed, task_removed_cb, ui);
	pan_callback_remove (queue_task_moved, task_moved_cb, ui);

	g_source_remove (ui->timeout_id);

	pan_warn_if_fail (ui == manager_ui);
	g_free (manager_ui);
	manager_ui = NULL;
}

/**
***
**/

static ManagerUI*
task_manager_build_ui (GtkWidget * window)
{
	//gboolean b;
	ManagerUI * ui = g_new0 (ManagerUI, 1);
	GtkWidget * w;
	GtkWidget * toolbar;
	GtkWidget * vbox;
	GdkPixbuf * pixbuf;
	GtkIconSet * icon_set;
	GtkWidget * image;
	char* list_titles [5];

	ui->window = window;
	vbox = gtk_vbox_new (FALSE, GUI_PAD_SMALL);

	/* button bar */
	toolbar = gtk_toolbar_new ();

	/* up */
	gtk_toolbar_insert_stock (GTK_TOOLBAR(toolbar), GTK_STOCK_GO_UP,
	                          _("Move Selected Task(s) Up"), NULL,
	                          G_CALLBACK(up_clicked_cb), ui, -1);
	/* top */
	gtk_toolbar_insert_stock (GTK_TOOLBAR(toolbar), GTK_STOCK_GOTO_TOP,
	                          _("Move Selected Task(s) to Top"), NULL,
	                          G_CALLBACK(top_clicked_cb), ui, -1);

	gtk_toolbar_insert_space (GTK_TOOLBAR(toolbar), -1);

	/* down */
	gtk_toolbar_insert_stock (GTK_TOOLBAR(toolbar), GTK_STOCK_GO_DOWN,
	                          _("Move Selected Task(s) Down"), NULL,
	                          G_CALLBACK(down_clicked_cb), ui, -1);
	/* bottom */
	gtk_toolbar_insert_stock (GTK_TOOLBAR(toolbar), GTK_STOCK_GOTO_BOTTOM,
	                          _("Move Selected Task(s) to Bottom"), NULL,
	                          G_CALLBACK(bottom_clicked_cb), ui, -1);

	gtk_toolbar_insert_space (GTK_TOOLBAR(toolbar), -1);

	/* redo */
	gtk_toolbar_insert_stock (GTK_TOOLBAR(toolbar), GTK_STOCK_REDO,
	                          _("Requeue selected failed Task(s)"), NULL,
	                          G_CALLBACK(requeue_cb), ui, -1);

	gtk_toolbar_insert_space (GTK_TOOLBAR(toolbar), -1);

	/* pause button */
	image = gtk_image_new_from_stock (GTK_STOCK_STOP, GTK_ICON_SIZE_SMALL_TOOLBAR);
	w = gtk_toolbar_append_element (GTK_TOOLBAR(toolbar),
	                                GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL,
	                                _("Pause"), _("Pause the Task Manager"), NULL,
	                                image, G_CALLBACK(pause_toggled_cb), ui);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(w), queue_is_paused());

        /* edit per-server maximum */
	pixbuf = gdk_pixbuf_new_from_inline (-1, icon_server, FALSE, NULL);
	icon_set = gtk_icon_set_new_from_pixbuf (pixbuf);
	g_object_unref (G_OBJECT(pixbuf));
	image = gtk_image_new_from_icon_set (icon_set, GTK_ICON_SIZE_SMALL_TOOLBAR);
	gtk_icon_set_unref (icon_set);
	gtk_toolbar_append_element (GTK_TOOLBAR(toolbar),
	                            GTK_TOOLBAR_CHILD_BUTTON, NULL,
				    _("Server"), _("Set Per-Server Connection Limits"), NULL,
				    image, G_CALLBACK(prefs_spawn_to_news_connections), ui);

	gtk_toolbar_insert_space (GTK_TOOLBAR(toolbar), -1);

        /* cancel */
	gtk_toolbar_insert_stock (GTK_TOOLBAR(toolbar), GTK_STOCK_CANCEL,
	                          _("Cancel Selected Task(s)"), NULL,
	                          G_CALLBACK(cancel_clicked_cb), ui, -1);

	/* clear */
	gtk_toolbar_insert_stock (GTK_TOOLBAR(toolbar), GTK_STOCK_CLEAR,
	                          _("Cancel All Task(s)"), NULL,
	                          G_CALLBACK(clear_queue_clicked_cb), ui, -1);

	/* add button bar to the queued window */
	gtk_box_pack_start (GTK_BOX(vbox), toolbar, FALSE, FALSE, 0);

        /* queue list */
        list_titles[0] = _("Status");
	list_titles[1] = _("% Done");
	list_titles[2] = _("Server");
        list_titles[3] = _("Transfer Rate");
	list_titles[4] = _("Description");
	ui->list = w = gtk_clist_new_with_titles (5, list_titles);
	gtk_clist_set_column_width (GTK_CLIST(w), 0, 200);
	gtk_clist_set_column_width (GTK_CLIST(w), 4, 1500);
	gui_restore_column_widths (ui->list, "task_manager_2");

	gtk_clist_set_selection_mode (GTK_CLIST(w), GTK_SELECTION_MULTIPLE);
	ui->moved_signal_id = g_signal_connect (w, "row_move", G_CALLBACK(queue_row_move_cb), ui);
	w = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(w), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (w), ui->list);
	gtk_box_pack_start (GTK_BOX(vbox), w, TRUE, TRUE, 0);

	pan_callback_add (queue_task_added, task_added_cb, ui);
	pan_callback_add (queue_task_removed, task_removed_cb, ui);
	pan_callback_add (queue_task_moved, task_moved_cb, ui);

	ui->titlebar_needs_refresh = TRUE;
	ui->timeout_id = pan_timeout_add (2000, periodic_update_timer, ui);
	ui->top = vbox;
        g_signal_connect (ui->top, "destroy", G_CALLBACK(ui_destroy_cb), ui);
	return ui;
}

/**
***  Titlebar dirty
**/

static void
update_titlebar_cb (gpointer call_obj, gpointer call_arg, gpointer user_data)
{
	((ManagerUI*)(user_data))->titlebar_needs_refresh = TRUE;
}
static void
remove_update_titlebar_callbacks_cb (GtkObject * o, gpointer user_data)
{
	pan_callback_remove (queue_task_added, update_titlebar_cb, user_data);
	pan_callback_remove (queue_task_removed, update_titlebar_cb, user_data);
	pan_callback_remove (queue_task_status_changed, update_titlebar_cb, user_data);
}

/************
*************  PUBLIC ROUTINES
************/

void
task_manager_spawn (void)
{
	/* There can be only one */
	if (manager_ui != NULL)
	{
		pan_lock();
		if (!GTK_WIDGET_MAPPED(GTK_WIDGET(manager_ui->window)))
			gtk_widget_map ( GTK_WIDGET(manager_ui->window));
		gdk_window_raise (manager_ui->window->window);
		gdk_window_show (manager_ui->window->window);
		pan_unlock();
		return;
	}

	manager_ui = task_manager_build_ui (gtk_window_new(GTK_WINDOW_TOPLEVEL));
	ui_populate (manager_ui);

	pan_callback_add (queue_task_added, update_titlebar_cb, manager_ui);
	pan_callback_add (queue_task_removed, update_titlebar_cb, manager_ui);
	pan_callback_add (queue_task_status_changed, update_titlebar_cb, manager_ui);
	g_signal_connect (manager_ui->window, "delete_event",
	                  G_CALLBACK(window_delete_event_cb), manager_ui);
	g_signal_connect (manager_ui->window, "destroy",
	                  G_CALLBACK(remove_update_titlebar_callbacks_cb), manager_ui);
	g_signal_connect (manager_ui->window, "key_press_event",
	                  G_CALLBACK(gui_key_press_cb), manager_ui);
	gtk_window_set_title (GTK_WINDOW(manager_ui->window), _("Pan - Task Manager"));
	gtk_container_add (GTK_CONTAINER(manager_ui->window), manager_ui->top);

	gtk_window_set_default_size (GTK_WINDOW (manager_ui->window), 600, 400);
	gui_restore_window_size (manager_ui->window, "task_manager_2");
	gtk_widget_show_all (manager_ui->window);
}

